<?php

declare(strict_types=1);

namespace Drupal\Tests\data_pipelines_elasticsearch\Kernel;

use Drupal\Core\Messenger\MessengerInterface;
use Drupal\Tests\data_pipelines\Kernel\DatasetKernelTestBase;
use Drupal\Tests\data_pipelines_elasticsearch\Traits\ElasticsearchTestDestinationTrait;
use Drupal\data_pipelines\DatasetData;
use Drupal\data_pipelines\DatasetPipelinePluginManager;
use Drupal\data_pipelines\Entity\DestinationInterface;
use Drupal\data_pipelines_elasticsearch\Plugin\DatasetDestination\ElasticSearchDestination;
use Elasticsearch\Client;
use Prophecy\Argument;
use Prophecy\PhpUnit\ProphecyTrait;
use Psr\Log\NullLogger;
use Symfony\Component\ErrorHandler\BufferingLogger;

/**
 * Defines a class for testing elastic indexer.
 *
 * @group data_pipelines
 *
 * @covers \Drupal\data_pipelines_elasticsearch\Plugin\DatasetDestination\ElasticSearchDestination
 */
class ElasticsearchDestinationTest extends DatasetKernelTestBase {

  use ProphecyTrait;
  use ElasticsearchTestDestinationTrait;

  /**
   * Tests index creation.
   */
  public function testIndexCreation() {
    $dataset = $this->createTestDataset();
    $indices = $this->prophesizeIndices(FALSE);

    $indices->create([
      'index' => self::INDEX_PREFIX . $dataset->getMachineName(),
      'body' => [
        'settings' => [
          'number_of_shards' => ElasticSearchDestination::INDEX_SHARDS,
          'number_of_replicas' => ElasticSearchDestination::INDEX_REPLICAS,
          'codec' => ElasticSearchDestination::INDEX_CODEC,
        ],
      ],
    ])->shouldBeCalled();
    $client = $this->prophesize(Client::class);
    $client->indices()->willReturn($indices->reveal());

    $client->bulk(Argument::any())->willReturn([]);

    $pipeline_manager = $this->prophesize(DatasetPipelinePluginManager::class);
    $pipeline_manager->hasDefinition(Argument::exact($dataset->getPipelineId()))->willReturn(TRUE);
    $pipeline_manager->getDefinition(Argument::exact($dataset->getPipelineId()))->willReturn([]);

    $indexer = $this->mockIndexer($pipeline_manager->reveal(), new NullLogger(), $client->reveal());
    $this->assertTrue($indexer->beginProcessing($dataset));
  }

  /**
   * Tests index not created if exists.
   */
  public function testIndexNotCreatedIfExists() {
    $dataset = $this->createTestDataset();
    $indices = $this->prophesizeIndices(TRUE);
    $indices->create(Argument::any())->shouldNotBeCalled();
    $client = $this->prophesize(Client::class);
    $client->indices()->willReturn($indices->reveal());

    $client->bulk(Argument::any())->willReturn([]);

    $pipeline_manager = $this->prophesize(DatasetPipelinePluginManager::class);
    $pipeline_manager->hasDefinition(Argument::exact($dataset->getPipelineId()))->willReturn(TRUE);
    $pipeline_manager->getDefinition(Argument::exact($dataset->getPipelineId()))->willReturn([]);

    $indexer = $this->mockIndexer($pipeline_manager->reveal(), new NullLogger(), $client->reveal());
    $this->assertTrue($indexer->beginProcessing($dataset));
  }

  /**
   * Tests index existence failure.
   */
  public function testIndexExistenceFailure() {
    $dataset = $this->createTestDataset();
    $indices = $this->prophesizeIndices(TRUE);
    $indices->exists(Argument::any())->willThrow(new \Exception('Barf'));
    $indices->create(Argument::any())->shouldNotBeCalled();
    $client = $this->prophesize(Client::class);
    $client->indices()->willReturn($indices->reveal());

    $pipeline_manager = $this->prophesize(DatasetPipelinePluginManager::class);
    $pipeline_manager->hasDefinition(Argument::exact($dataset->getPipelineId()))->willReturn(TRUE);
    $pipeline_manager->getDefinition(Argument::exact($dataset->getPipelineId()))->willReturn([]);

    $indexer = $this->mockIndexer($pipeline_manager->reveal(), new NullLogger(), $client->reveal());
    $this->assertFalse($indexer->beginProcessing($dataset));
  }

  /**
   * Tests index creation failure.
   */
  public function testIndexCreationFailure() {
    $dataset = $this->createTestDataset();
    $indices = $this->prophesizeIndices(FALSE);
    $indices->create(Argument::any())->willThrow(new \Exception('Barf'));
    $client = $this->prophesize(Client::class);
    $client->indices()->willReturn($indices->reveal());

    $pipeline_manager = $this->prophesize(DatasetPipelinePluginManager::class);
    $pipeline_manager->hasDefinition(Argument::exact($dataset->getPipelineId()))->willReturn(TRUE);
    $pipeline_manager->getDefinition(Argument::exact($dataset->getPipelineId()))->willReturn([]);

    $indexer = $this->mockIndexer($pipeline_manager->reveal(), new NullLogger(), $client->reveal());
    $dataset = $this->createTestDataset();
    $this->assertFalse($indexer->beginProcessing($dataset));
  }

  /**
   * Tests indexing.
   */
  public function testIndexing() {
    $dataset = $this->createTestDataset();
    $indices = $this->prophesizeIndices(TRUE);
    $client = $this->prophesize(Client::class);
    $client->indices()->willReturn($indices->reveal());

    $pipeline_manager = $this->prophesize(DatasetPipelinePluginManager::class);
    $pipeline_manager->hasDefinition(Argument::exact($dataset->getPipelineId()))->willReturn(TRUE);
    $pipeline_manager->getDefinition(Argument::exact($dataset->getPipelineId()))->willReturn([]);

    $bulk = [
      'index' => self::INDEX_PREFIX . $dataset->getMachineName(),
      'body' => [],
    ];
    foreach ($dataset->getDataIterator() as $delta => $data) {
      assert($data instanceof DatasetData);
      $bulk['body'][] = [
        'index' => [
          '_id' => $dataset->getMachineName() . ':' . $delta,
        ],
      ];
      $bulk['body'][] = ['@delta' => $delta] + $data->getArrayCopy();
      if (count($bulk['body']) === 200) {
        $client->bulk($bulk)->shouldBeCalled();
        $bulk['body'] = [];
      }
    }
    if (count($bulk['body']) > 0) {
      $client->bulk($bulk)->shouldBeCalled();
    }

    $indexer = $this->mockIndexer($pipeline_manager->reveal(), new NullLogger(), $client->reveal());
    $this->assertTrue($indexer->beginProcessing($dataset));
    $this->assertTrue($indexer->processChunk($dataset, iterator_to_array($dataset->getDataIterator())));
  }

  /**
   * Tests deletion.
   */
  public function testDeleting() {
    $dataset = $this->createTestDataset();
    $destination = $this->prophesize(DestinationInterface::class);
    $indices = $this->prophesizeIndices(TRUE);
    $client = $this->prophesize(Client::class);
    $client->indices()->willReturn($indices->reveal());

    $pipeline_manager = $this->prophesize(DatasetPipelinePluginManager::class);
    $pipeline_manager->hasDefinition(Argument::exact($dataset->getPipelineId()))->willReturn(TRUE);
    $pipeline_manager->getDefinition(Argument::exact($dataset->getPipelineId()))->willReturn([]);

    $expected = [
      'index' => self::INDEX_PREFIX . $dataset->getMachineName(),
    ];
    $indices->delete($expected)->shouldBeCalled();

    $indexer = $this->mockIndexer($pipeline_manager->reveal(), new NullLogger(), $client->reveal());
    $this->assertTrue($indexer->deleteDataSet($dataset, $destination->reveal()));
  }

  /**
   * Tests deletion failure.
   */
  public function testDeletingException() {
    $dataset = $this->createTestDataset();
    $destination = $this->prophesize(DestinationInterface::class);
    $indices = $this->prophesizeIndices(TRUE);
    $client = $this->prophesize(Client::class);
    $client->indices()->willReturn($indices->reveal());

    $pipeline_manager = $this->prophesize(DatasetPipelinePluginManager::class);
    $pipeline_manager->hasDefinition(Argument::exact($dataset->getPipelineId()))->willReturn(TRUE);
    $pipeline_manager->getDefinition(Argument::exact($dataset->getPipelineId()))->willReturn([]);
    $messenger = $this->prophesize(MessengerInterface::class);

    $expected = [
      'index' => self::INDEX_PREFIX . $dataset->getMachineName(),
    ];
    $exception = new \Exception('barf');
    $indices->delete($expected)->willThrow($exception);

    $logger = new BufferingLogger();
    $indexer = $this->mockIndexer($pipeline_manager->reveal(), $logger, $client->reveal());
    $indexer->setMessenger($messenger->reveal());

    $indexer->deleteDataSet($dataset, $destination->reveal());
    $logs = $logger->cleanLogs();
    $this->assertCount(1, $logs);
    [, $message] = array_pop($logs);
    $this->assertEquals(sprintf('The dataset %s could not be removed from the index due to %s', $dataset->label(), $exception->getMessage()), $message);
  }

}
